(* $Id: gontcodegen.ml,v 1.9 2001/11/16 00:00:46 malekith Exp $ *)

(* This file generally converts Gont.* -> Ksi.*. *)

open Gontutil
open Gontbind
open Gonttypes

let ksi_runtime_proto () = 
  let fnc (r, n, a) = Ksi.FuncProto([Ksi.Extern], r, n, a) in
  let ptr = Ksi.Ptr(Ksi.Void) in
  let size_t = Ksi.uptroff () in
[
  Ksi.TopComment("BEG Runtime proto");
  Ksi.StructDef("lambda_exp$", [], [(ptr, "fnc"); (ptr, "closure")]);
  fnc(ptr, "make_closure$", [size_t]);
  fnc(Ksi.Ptr(Ksi.Struct("lambda_exp$")), "make_lambda$", [ptr; ptr]);
  Ksi.Global([Ksi.Extern], ptr, "global_closure$");
  Ksi.TopComment("END Runtime proto")
]
;;

(* This is going to prepend module name, mangle name or whatever. *)
let ksi_name s = s

let rec ksi_type t =
  match t with
      Int | Bool -> Ksi.ptroff ()
    | Void -> Ksi.Void
    | Float -> Ksi.c_type(Ksi.C_double)
(*  | RealFunc(r, a) -> Ksi.FuncPtr(ksi_type r, (List.map ksi_type a)) *)
    | Func(_, _) -> Ksi.Ptr(Ksi.Struct("lambda_exp$"))
    | String | TyVar(_) -> Ksi.Ptr(Ksi.Void)

let ksi_storage s =
  match s with
      Local -> [Ksi.Local]
    | Public -> [Ksi.Public]
;;

type fnc_ctx = {ctx_level: int;
                ctx_closure_type: id;
                ctx_closure_name: id;
                ctx_aux_decls: Ksi.decl list;
                ctx_storage: storage}
;;

let ksi_op2 o =
  match o with
      IntPlus | FloatPlus -> Ksi.Plus
    | IntMinus | FloatMinus -> Ksi.Minus
    | IntMult | FloatMult -> Ksi.Mult
    | IntDiv -> Ksi.TruncDiv
    | IntMod -> Ksi.TruncMod
    | FloatDiv -> Ksi.RDiv
    | Assign -> assert false
  
let ksi_op1 o =
  match o with
      IntNegate | FloatNegate -> Ksi.Negate
;;

let rec ksi_stmt (c, s) =
  match s.raw_stmt with
      Skip ->
        (c, Ksi.Nop)
    | Exp(e) ->
        ksi_exp (c, e)
    | Return(e) ->
        let (c', e') = ksi_exp (c, e) in 
        (c', Ksi.Return(e'))
    | ReturnVoid ->
        (c, Ksi.Return(Ksi.Nop))
    | Block([]) ->
        (c, Ksi.Nop)
    | Block([s]) ->
        let (c', s') = ksi_stmt (c, s) in (c', Ksi.Bind(s'))
    | Block([s1; s2]) ->
        let (c1, s1') = ksi_stmt (c, s1) in
        let (c2, s2') = ksi_stmt (c1, s2) in
        (c2, Ksi.Bind(Ksi.Compound(s1', s2')))
    | Block(s) ->
        let (c', s') = list_apply ksi_stmt (c, s) in
        (c', Ksi.Bind(Ksi.CompoundList(s')))
    | VarDef v ->
        (match v.var_symbol with
            Some sym -> 
              if sym#in_closure then
                (c, Ksi.Comment(v.var_name ^ " in closure", Ksi.Nop))
              else
                (c, Ksi.Var(ksi_type v.var_type, v.var_name))
          | None -> assert false)
        
and ksi_exp (c, e) =
  let toplev_func sym =
    match sym#get_type with
        Func(_, _) when sym#get_scope_level == 0 -> true
      | _ -> false in
  match e.raw_exp with
      SimpleBinOp(Assign, e1, e2) ->
        let (c1, e1') = ksi_exp (c, e1) in
        let (c2, e2') = ksi_exp (c1, e2) in
        (c2, Ksi.Modify(e1', e2'))
    | SimpleBinOp(op, e1, e2) -> 
        let (c1, e1') = ksi_exp (c, e1) in
        let (c2, e2') = ksi_exp (c1, e2) in
        (c2, Ksi.Binary(ksi_op2 op, e1', e2'))
    | SimpleUnOp(op, e) -> 
        let (c', e') = ksi_exp (c, e) in
        (c', Ksi.Unary(ksi_op1 op, e'))
    | Ref ({ref_symbol = Some sym}) when (toplev_func sym) ->
      (* Need to encapsulate toplevel function into a function with closure. *)
      let aux_name = (seq_name "closurized") ^ "$" in
      let (ret_type, parm_types) = 
        match sym#get_type with
            Func(r, p) -> (r, p)
          | _ -> assert false in
      let cnt = ref 0 in
      let parm_names = 
        (List.map (fun _ -> incr cnt; 
                            ("_" ^ string_of_int(!cnt))) parm_types) in
      let aux = 
         Ksi.Func([Ksi.Local], ksi_type ret_type, aux_name, 
                  (Ksi.Ptr(Ksi.Void), "dummy_cl$") :: 
                    (List.map2 (fun t n -> (ksi_type t, n)) parm_types parm_names),
                  Ksi.Return(Ksi.Call(sym#get_name, 
                    (List.map (fun n -> Ksi.Ref(n)) parm_names)))) in
      let lambda_val = 
         Ksi.Call("make_lambda$", 
                  [Ksi.Convert(Ksi.Ptr(Ksi.Void), 
                               Ksi.FuncRef(aux_name));
                   Ksi.Ref(c.ctx_closure_name)]) in
      ({c with ctx_aux_decls = aux :: c.ctx_aux_decls}, lambda_val)
    | Ref ({ref_symbol = Some sym} as r) ->
        let code = 
          let rec refer lev =
            if lev == 1 then 
              Ksi.Ref("parent_cl$")
            else 
               Ksi.Arrow(refer (lev-1), "parent_cl$") in
          if r.ref_level == 0 then
            if sym#in_closure then
              Ksi.Arrow(Ksi.Ref(c.ctx_closure_name), 
                                         sym#get_name)
            else
              Ksi.Ref(sym#get_name)
          else
            (assert sym#in_closure;
             Ksi.Arrow(refer r.ref_level, sym#get_name)) in
        (c, code)
    | Ref _ -> assert false
    (* The easy case -- global function call. *)
    | Call({raw_exp = Ref {ref_symbol = Some sym}}, p) when (toplev_func sym) ->
        let (c', p') = list_apply ksi_exp (c, p) in
        (c', Ksi.Call(sym#get_name, p'))
    (* The hard case -- function call with closure. *)
    | Call(e, p) ->
        let (c', e') = ksi_exp (c, e) in
        let (c'', p') = list_apply ksi_exp (c', p) in
            (* cast might be required, for example when this is return
             * value from polimorphic function returing 'a, but we leave
             * it for now. XXX
            Ksi.Convert(Ksi.Ptr(Ksi.Struct("lambda_exp$")), e') *)
        let lambda_fld n = Ksi.Arrow(e', n) in
        let (ret_type, parm_types) =
          match type_of e with
              Func(r, p) -> (r, p)
            | _ -> assert false in
        let fnc_ptr = Ksi.Convert(Ksi.FuncPtr(ksi_type ret_type,
                                              Ksi.Ptr(Ksi.Void) :: 
                                              (List.map ksi_type parm_types)),
                                  lambda_fld "fnc") in
        (c'', Ksi.PtrCall(fnc_ptr, (lambda_fld "closure") :: p'))
    | IntConst(i) -> (c, Ksi.IntConst(ksi_type Int, i))
    | BoolConst(b) -> (c, Ksi.BoolConst(b))
    | FloatConst(f) -> (c, Ksi.FloatConst(ksi_type Float, f))
    | StringConst(s) -> (c, Ksi.StringConst(s))
    | Lambda(p, t, s) ->
      let aux_name = (seq_name "lambda") ^ "$" in
      let aux_decls = ksi_fnc_decl c (t, p) aux_name s in
      let lambda_val = 
        Ksi.Call("make_lambda$", 
                 [Ksi.Convert(Ksi.Ptr(Ksi.Void), 
                              Ksi.FuncRef(aux_name));
                  Ksi.Ref(c.ctx_closure_name)]) in
      ({c with ctx_aux_decls = aux_decls @ c.ctx_aux_decls}, lambda_val)
  
and ksi_fnc_decl c1 (ret_type, parms) fnname body =
  let closure_struct name cl =
    let flds = (List.map (fun sym -> (ksi_type sym#get_type, 
                                      sym#get_name)) cl) in
    match c1.ctx_closure_type with
        "" -> Ksi.StructDef(name, [], flds)
      | _ -> let parent_cl = (Ksi.Ptr(Ksi.Struct(c1.ctx_closure_type)), 
                              "parent_cl$") in
             Ksi.StructDef(name, [], parent_cl :: flds) in
  let (need_cl, cl) = closure (c1.ctx_level + 1) body in
  let c2 = {ctx_level = c1.ctx_level + 1;
            ctx_closure_name = "global_closure$";
            ctx_closure_type = "";
            ctx_aux_decls = [];
            ctx_storage = Local} in
  let c3 = if not need_cl then c2 else
           let name = (seq_name "closure") ^ "_t$" in
               {c2 with
                ctx_closure_name = "cl$";
                ctx_closure_type = name;
                ctx_aux_decls = (closure_struct name cl) :: c2.ctx_aux_decls} in
  let (c4, code) = ksi_stmt (c3, body) in
  let code_with_cl =
    if not need_cl then 
      code 
    else
      let cl_struct = Ksi.Struct(c3.ctx_closure_type) in
      let cl_var = Ksi.Var(Ksi.Ptr(cl_struct), c3.ctx_closure_name) in
      let cl_init = 
          Ksi.Modify(Ksi.Ref(c3.ctx_closure_name),
                     Ksi.Convert(Ksi.Ptr(cl_struct),
                                 Ksi.Call("make_closure$",
                                          [Ksi.SizeOf(cl_struct)]))) in
      let cl_copy_parent =
        match c1.ctx_closure_type with
            "" -> Ksi.Nop
          | _ -> Ksi.Modify(Ksi.Arrow(Ksi.Ref(c3.ctx_closure_name), 
                                              "parent_cl$"),
                            Ksi.Ref("parent_cl$")) in
      Ksi.CompoundList([cl_var; cl_init; cl_copy_parent; code]) in
  let (ksi_ret_type, ksi_parms) = (ksi_type ret_type, 
                                   (List.map (fun (t, n) -> ksi_type t, n)
                                             parms)) in
  let ksi_parms' =
    if c1.ctx_closure_type <> "" then
      (Ksi.Ptr(Ksi.Struct(c1.ctx_closure_type)), "parent_cl$") :: ksi_parms
    else if c1.ctx_level > 0 then
      (Ksi.Ptr(Ksi.Void), "dummy_cl$") :: ksi_parms
    else
      ksi_parms in
  let func_decl = 
    Ksi.Func(ksi_storage c1.ctx_storage,
             ksi_ret_type,
             fnname,
             ksi_parms',
             code_with_cl) in
  match c4.ctx_aux_decls with
      [] -> 
        [Ksi.TopComment("END " ^ fnname); 
         func_decl; 
         Ksi.TopComment("BEG " ^ fnname)]
    | aux -> 
        Ksi.TopComment("END " ^ fnname) ::
        func_decl ::
        Ksi.TopComment("BEG " ^ fnname) :: 
        aux @ 
        [Ksi.TopComment("AUX " ^ fnname)]
;;

let global_ctx = {ctx_level = 0; 
                  ctx_closure_name = "global_undef$";
                  ctx_closure_type = "";
                  ctx_aux_decls = [];
                  ctx_storage = Local}
;;

let ksi_code d =
  match d with
      FuncDecl(f) ->
         List.rev
            (ksi_fnc_decl {global_ctx with ctx_storage = f.fn_storage}
                          (f.fn_ret_type, f.fn_parms) 
                          f.fn_name 
                          f.fn_body)
    | GlobalDecl(s, t, n, _) ->
         [Ksi.Global(ksi_storage s,
                     ksi_type t,
                     ksi_name n)]
    | ExternC(x) ->
         [Ksi.FuncProto([Ksi.Extern], 
                        ksi_type x.xc_ret_type,
                        x.xc_name,
                        (List.map ksi_type x.xc_parms))]
;;
